from .data import *
from .world import *
from .tools import *
from .pvslice import *
from .spatial import *
from .spectral import *
from .logging import get_logger
from .lines import *

'''
A potentially simple module for handling long-slit and pseudo-long-slit data
'''

#if __name__ == '__main__':#
#    pass
#else:
#    pass

get_logger()
